<?php
/*
NOTICE OF LICENSE

This source file is subject to the NekloEULA that is bundled with this package in the file ICENSE.txt.

It is also available through the world-wide-web at this URL: http://store.neklo.com/LICENSE.txt

Copyright (c)  Neklo (http://store.neklo.com/)
*/


namespace Neklo\Scheduler\Observer;

use Exception;
use Magento\Framework\App\ObjectManager;
use Magento\Framework\Event\Observer;
use Neklo\Scheduler\Api\Data\JobInterface;
use Neklo\Scheduler\Model\Job;
use Neklo\Scheduler\Model\JobFactory;
use Psr\Log\LoggerInterface;
use Magento\Framework\Lock\LockManagerInterface;

/**
 * Class ProcessCronQueueObserver
 * @package Neklo\Scheduler\Observer
 */
class ProcessCronQueueObserver extends \Magento\Cron\Observer\ProcessCronQueueObserver
{
    /**
     * @var string
     */
    const FILOBLU_NEKLO_CRON_LOCK = 'FILOBLU_NEKLO_CRON_LOCK';

    /**
     * Process cron queue
     * Generate tasks schedule
     * Cleanup tasks schedule
     *
     * @param Observer $observer
     * @return void
     * @SuppressWarnings(PHPMD.CyclomaticComplexity)
     * @SuppressWarnings(PHPMD.NPathComplexity)
     * @SuppressWarnings(PHPMD.UnusedFormalParameter)
     */
    public function execute(Observer $observer)
    {
        parent::execute($observer);
        $this->updateJobs();
    }

    /**
     * Update Neklo Cron Jobs
     */
    private function updateJobs()
    {
        // Get only job to insert
        $jobsToInsert = $this->getJobToInsert();

        // If there are no job to insert return
        if(empty($jobsToInsert)){
            return;
        }

        $objectManager = ObjectManager::getInstance();

        /** @var LoggerInterface $logger */
        $logger = $objectManager->create(LoggerInterface::class);

        /** @var LockManagerInterface $lockManager */
        $lockManager = $objectManager->create(LockManagerInterface::class);

        //Check if table is locked
        if($lockManager->isLocked(self::FILOBLU_NEKLO_CRON_LOCK)) {
            return;
        }

        //Lock the table to avoid job duplication
        //TODO: SET TIMEOUT FROM SETTINGS
        if(!$lockManager->lock(self::FILOBLU_NEKLO_CRON_LOCK,60)) {
            $logger->warning('There was an error when trying to lock '.self::FILOBLU_NEKLO_CRON_LOCK.'. Skipping');
            return;
        }

        try {
            /** @var JobFactory $jobGroups */
            //$jobGroups = $this->_config->getJobs();

            foreach ($jobsToInsert as $group => $jobs) {
                foreach ($jobs as $jobName => $job) {
                    try {
                        /** @var Job $model */
                        $model = $objectManager->create(Job::class);
                        //$model = $model->load($jobName, JobInterface::JOB_NAME);
                        //if (!$model->getId()) {
                            $model->setName($jobName)
                                ->setMethod($job[JobInterface::JOB_METHOD])
                                ->setInstance($job[JobInterface::JOB_CLASS])
                                ->setEnabled(false)
                                ->setGroup($group);
                            if (isset($job[JobInterface::JOB_SCHEDULE])) {
                                $model->setSchedule($job[JobInterface::JOB_SCHEDULE]);
                            }
                            else {
                                // IF THERE IS NO SCHEDULE WE SET DEFAULT FREQUENCY
                                $model->setSchedule('* * * * *');
                            }

                            if (isset($job[JobInterface::JOB_CONFIG])) {
                                $model->setConfigPath($job[JobInterface::JOB_CONFIG]);
                            }

                            $model->save();
                        //}
                    } catch (Exception $e) {
                        $logger->error($e->getMessage(), ['exception' => $e]);
                    }
                }
            }
        }
        catch (\Exception $e){
            $logger->error($e->getMessage(), ['exception' => $e]);
        }
        finally {
            // Unlock the table
            for($wait = 0; ($status = $lockManager->unlock(self::FILOBLU_NEKLO_CRON_LOCK)) && $wait < 10; $wait++){
                sleep($wait*10);
            }
            if($status){
                return;
            }
            // If can't unlock the trigger alert administrators to manually unlock
            $logger->alert('Unable to release neklo_cron table throught lock manager with name '.self::FILOBLU_NEKLO_CRON_LOCK);
        }

    }

    /**
     * Generate Jobs
     *
     * @param array $jobs
     * @param array $exists
     * @param string $groupId
     */
    protected function _generateJobs($jobs, $exists, $groupId)
    {
        $jobsToSchedule = [];

        foreach ($jobs as $jobCode => $jobConfig) {

            if (!isset($jobConfig[JobInterface::JOB_ENABLED])) {
                continue;
            }

            if (false === (bool)$jobConfig[JobInterface::JOB_ENABLED]) {
                continue;
            }

            $jobsToSchedule[$jobCode] = $jobConfig;
        }

        parent::_generateJobs($jobsToSchedule, $exists, $groupId);
    }

    /**
     * Get jobs to insert in neklo cron table
     *
     * @return array
     */
    protected function getJobToInsert(){

        // Get all jobs from magento config
        $configJobs = [];
        $jobGroups = $this->_config->getJobs();
        foreach ($jobGroups as $group => $jobs){
            foreach ($jobs as $jobName => $job) {
                $configJobs[$jobName] = $job;
                $configJobs[$jobName][JobInterface::JOB_GROUP] = $group;
            }
        }

        // Get all job from neklo_cron table
        $tableJobs = [];
        $objectManager = \Magento\Framework\App\ObjectManager::getInstance(); // Instance of object manager
        $resource = $objectManager->create('Magento\Framework\App\ResourceConnection');
        $connection = $resource->getConnection();
        $nekloCronTable = $resource->getTableName('neklo_cron');
        $cronJobs = $connection->fetchAll("SELECT name FROM $nekloCronTable", \Zend_Db::FETCH_ASSOC);

        foreach($cronJobs as $cron) {
            $tableJobs[] = $cron['name'];
        }

        // Get difference between arrays
        $diff = array_diff(array_keys($configJobs),$tableJobs);

        if(empty($diff)){
            return [];
        }

        $result = [];
        foreach($diff as $name){
            $result[$configJobs[$name][JobInterface::JOB_GROUP]][$name] = $configJobs[$name];
        }

        return $result;

    }
}
